@@ -1,6 +1,6 @@
 /*
     Configuration management header file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -18,20 +18,24 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 
-#ifndef __CONFIG_H
-#define __CONFIG_H
+#ifndef _CONFIG_H
+#define _CONFIG_H
 
 
 #include <stdio.h>
 
 
-#define MAXLINE		1024
+#define MAXLINE		1024	/* change this to get more directory entries */
 #define NO_SEEK		0
 #define DO_SEEK		1
 
 #define STATUS_BAD	0
 #define STATUS_OK	1
 
+#define IFS  		';'	/* internal field separator */
+#define ICS		'#'	/* internal comment separator */
+#define IAS		'='	/* internal assignment separator */
+
 
 typedef struct
 {
@@ -51,4 +55,4 @@ void configuration_getfielddata(configuration *this, int sectionptr,
 				int fields, int seek);
 
 
-#endif
+#endif	/* _CONFIG_H */
