@@ -1,6 +1,6 @@
-/*  
+/*
     Panels management file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -41,6 +41,7 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 #include "window.h"
 #include "status.h"
 #include "panel.h"
+#include "fnmatch.h"
 
 
 /* This is really ugly ... */
@@ -147,13 +148,29 @@ char FileDisplayMode[FILE_DISPLAY_MODES][15] =
     "DateTime",
     "Size",
     "Mode",
-    "FullName"
+    "FullName",
+};
+
+
+#define FILE_SORT_METHODS	9
+
+char FileSortMethod[FILE_SORT_METHODS][15] = 
+{
+    "Name",
+    "Extension",
+    "Size",
+    "Date",
+    "Mode",
+    "OwnerId",
+    "GroupId",
+    "OwnerName",
+    "GroupName",
 };
 
 
 #define PANEL_FIELDS	17
 
-static char PanelFields[PANEL_FIELDS][40] = 
+static char PanelFields[PANEL_FIELDS][36] =
 {
     "PanelFrame",
     "PanelBackground",
@@ -178,7 +195,7 @@ static char PanelFields[PANEL_FIELDS][40] =
 static int PanelColors[PANEL_FIELDS] = 
 {
     WHITE, BLUE, YELLOW, ON, WHITE, ON, YELLOW, WHITE, 
-    CYAN, RED, OFF, RED, OFF, RED, OFF, BLACK, OFF
+    CYAN, RED, OFF, RED, OFF, RED, OFF, BLACK, OFF,
 };
 #else
 static int PanelColors[PANEL_FIELDS] = 
@@ -208,8 +225,11 @@ static int PanelColors[PANEL_FIELDS] =
 
 static int *UserHeartAttack;
 static int StartupFileDisplayMode;
+static int StartupFileSortMethod;
+static int CurrentSortMethod;
+static int LeadingDotMatch = OFF;
 static int InfoDisplay  = OFF;
-int FrameDisplay = OFF;		/* Sorry, I really nead it this way ... :-( */
+int FrameDisplay	= OFF;	/* Sorry, I really nead it this way ... :-( */
 
 int  tell(int);
 void uitclock(int);
@@ -246,10 +266,10 @@ panel *panel_init(int _lines, int _columns,
     this->begin_y = _begin_y;
     this->focus   = OFF;
     this->edt     = _edt;
-    this->info_dirty = ON;
     this->entries = this->selected_files = 0;
     this->last_index = -1;
-    this->mode = this->found_selection = this->end = 0;
+    this->display_mode = this->sort_method = 0;
+    this->found_selection = this->end = 0;
     this->current_entry = 0;
     this->first_on_screen = max(0, this->current_entry - (this->lines-2) + 1);
     this->on_screen = MAX_ENTRIES;
@@ -262,12 +282,15 @@ panel *panel_init(int _lines, int _columns,
 
     if (configured)
     {
-        this->mode = StartupFileDisplayMode;
+        this->display_mode = StartupFileDisplayMode;
+        this->sort_method  = StartupFileSortMethod;
         return this;
     }
+
     if (configuration_getstatus(config) == STATUS_OK)
     {
 	sectionptr = configuration_getsectionptr(config, "[UIT-Setup]");
+
         if (sectionptr != -1)
         {
 	    configuration_getfielddata(config, sectionptr,
@@ -275,26 +298,45 @@ panel *panel_init(int _lines, int _columns,
 				       &data, 1, DO_SEEK);
 	    if (data)
 	    {
-                for (i = 0; i < FILE_DISPLAY_MODES; i++)
-                    if (strcmp(data, FileDisplayMode[i]) == 0)
-                        break;
-                if (i == FILE_DISPLAY_MODES)
-                    fprintf(stderr, "Invald StartupFileDisplayMode (%s).\n",
+		for (i = 0; i < FILE_DISPLAY_MODES; i++)
+		    if (strcmp(data, FileDisplayMode[i]) == 0)
+			break;
+		if (i == FILE_DISPLAY_MODES)
+		    fprintf(stderr, "Invald StartupFileDisplayMode (%s).\n",
 			    data);
-                else
-                    this->mode = StartupFileDisplayMode = i;
+		else
+		    this->display_mode = StartupFileDisplayMode = i;
             }
+
+	    configuration_getfielddata(config, sectionptr,
+				       "StartupFileSortMethod",
+				       &data, 1, DO_SEEK);
+	    if (data)
+	    {
+		for (i = 0; i < FILE_SORT_METHODS; i++)
+		    if (strcmp(data, FileSortMethod[i]) == 0)
+			break;
+		if (i == FILE_SORT_METHODS)
+		    fprintf(stderr, "Invald StartupFileSortMethod (%s).\n",
+			    data);
+		else
+		    this->sort_method = StartupFileSortMethod = i;
+            }
+
 	    configuration_getfielddata(config, sectionptr, "InfoDisplay",
 				       &data, 1, DO_SEEK);
 	    InfoDisplay = data && strcmp(data, "ON") == 0;
 	    configuration_getfielddata(config, sectionptr, "FrameDisplay",
 				       &data, 1, DO_SEEK);
 	    FrameDisplay = data && strcmp(data, "ON") == 0;
+	    configuration_getfielddata(config, sectionptr, "LeadingDotMatch",
+				       &data, 1, DO_SEEK);
+	    LeadingDotMatch = data && strcmp(data, "ON") == 0;
 	}
 	
 	section = (LinuxConsole && ColorMonitor) ? cSection : bwSection;
-
 	sectionptr = configuration_getsectionptr(config, section);
+
         if (sectionptr != -1)
             for (i = 0; i < PANEL_FIELDS; i++)
             {
@@ -306,6 +348,7 @@ panel *panel_init(int _lines, int _columns,
                     PanelColors[i] = index;
             }
     }
+
     configured = 1;
     return this;
 }
@@ -343,25 +386,84 @@ char *cutname(char *name, int which, int how)
 static int sortfn(const void *_first, const void *_second)
 {
     int retval;
+    char *pfirst, *psecond;
     _dir_entry *first  = (_dir_entry *)_first;
     _dir_entry *second = (_dir_entry *)_second;
     int first_is_dir   = first->type  == DIR_ENTRY;
     int second_is_dir  = second->type == DIR_ENTRY;    
-    char *pfirst       = strrchr(first->name , '.');
-    char *psecond      = strrchr(second->name, '.');
 
-    if (first_is_dir == second_is_dir)
-        if (pfirst && psecond)
-	    return (retval = strcmp(++pfirst, ++psecond)) ?
-                    retval : strcmp(first->name, second->name);
-        else
-            return (pfirst || psecond) ?
-                   (pfirst ? -1 : 1) : strcmp(first->name, second->name);
-    else return first_is_dir ? -1 : 1;
+    if (first_is_dir != second_is_dir)
+	return first_is_dir ? -1 : 1;
+
+    switch (CurrentSortMethod)
+    {
+	case SORT_BY_NAME:
+
+	    pfirst = psecond = NULL;
+	    return strcmp(first->name, second->name);
+
+	case SORT_BY_EXTENSION:
+
+	    pfirst  = strrchr(first->name , '.');
+	    psecond = strrchr(second->name, '.');
+
+	    if (pfirst && psecond)
+		return (retval = strcmp(++pfirst, ++psecond)) ?
+		    retval : strcmp(first->name, second->name);
+	    else
+		return (pfirst || psecond) ?
+		       (pfirst ? -1 : 1) : strcmp(first->name, second->name);
+	    break;
+
+	case SORT_BY_SIZE:
+
+	    if (first->size == second->size)
+		return strcmp(first->name, second->name);
+	    return first->size - second->size;
+
+	case SORT_BY_DATE:
+
+	    if (first->mtime == second->mtime)
+		return strcmp(first->name, second->name);
+	    return first->mtime - second->mtime;
+
+	case SORT_BY_MODE:
+
+	    if (first->mode == second->mode)
+		return strcmp(first->name, second->name);
+	    return first->mode - second->mode;
+
+	case SORT_BY_OWNER_ID:
+
+	    if (first->uid == second->uid)
+		return strcmp(first->name, second->name);
+	    return first->uid - second->uid;
+
+	case SORT_BY_GROUP_ID:
+
+	    if (first->gid == second->gid)
+		return strcmp(first->name, second->name);
+	    return first->gid - second->gid;
+
+	case SORT_BY_OWNER_NAME:
+
+	    if (first->uid == second->uid)
+		return strcmp(first->name, second->name);
+	    return strcmp(first->owner, second->owner);
+
+	case SORT_BY_GROUP_NAME:
+
+	    if (first->gid == second->gid)
+		return strcmp(first->name, second->name);
+	    return strcmp(first->group, second->group);
+
+	default:
+	    fatal("bad sort method");
+    }
 }
 
 
-void  panel_nooptimizations(panel *this)
+void panel_nooptimizations(panel *this)
 {
     this->on_screen = MAX_ENTRIES;
 }
@@ -416,18 +518,19 @@ void panel_recover(panel *this)
 int panel_getdirinfo(panel *this, char *directory, int verify)
 {
     DIR *tmpdir;
-    struct dirent *d;
     struct stat s;
+    struct dirent *d;
 #ifdef HAVE_STATFS
     struct statfs fstat;
 #endif
-    struct passwd *pwd;
-    struct group *grp;
     struct tm *time;
-    int i, j, old_uid = -1, old_gid = -1, msdosfs = 0, namelen;
-    int old_entries = 0, backdir_index = -1, sz, hour;
+    struct group *grp;
+    struct passwd *pwd;
+    int dot_found = 0, dotdot_found = 0;
     _dir_entry *old_dir_entry = NULL, tmp;
     char old_path[PATH_MAX], temp[PATH_MAX];
+    int old_entries = 0, backdir_index = -1, sz, hour;
+    int i, j, old_uid = -1, old_gid = -1, msdosfs = 0, namelen;
 
 
     if (!(tmpdir = opendir(directory)))
@@ -460,18 +563,29 @@ int panel_getdirinfo(panel *this, char *directory, int verify)
         memset(this->dir_entry, 0, sizeof(this->dir_entry));
     }
 
-    this->selected_files = 0;
-    this->maxname = 0;
-    for (this->entries = 0;
-	 (d = readdir(this->dir)) && this->entries < MAX_ENTRIES;
+    for (this->selected_files = this->maxname = this->entries = 0;
+	 this->entries < MAX_ENTRIES && (d = readdir(this->dir));
 	 this->entries++)
     {
-       if (d->d_name[0] == '.' && !d->d_name[1])
-            { this->entries--; continue; }		/* ignore "." */
+        if (d->d_name[0] == '.' && !d->d_name[1])	/* ignore "." */
+	{
+	    this->entries--;
+	    dot_found = 1;
+	    continue;
+	}
+
         if (d->d_name[0] == '.' && d->d_name[1] == '.' && d->d_name[2] == 0)
+        {
+            dotdot_found = 1;
             if (this->path[1])
                 backdir_index = this->entries;
-            else { this->entries--; continue; }	/* ignore ".." if root dir */
+            else				/* ignore ".." if root dir */
+	    {
+		this->entries--;
+		continue;
+	    }
+        }
+
         s.st_ino = 0;
         stat(d->d_name, &s);
         this->dir_entry[this->entries].mode = s.st_mode;
@@ -561,7 +675,8 @@ int panel_getdirinfo(panel *this, char *directory, int verify)
 	    else
                 sprintf(this->dir_entry[this->entries].group, "%-7d",s.st_gid);
 	}
-	
+
+	this->dir_entry[this->entries].mtime = s.st_mtime;
 	time = localtime(&s.st_mtime);
 	if ((hour = time->tm_hour % 12) == 0) hour = 12;
 	sprintf(this->dir_entry[this->entries].date,"%2d-%02d-%02d %2d:%02d%c",
@@ -577,13 +692,25 @@ int panel_getdirinfo(panel *this, char *directory, int verify)
         this->maxname = max(this->maxname, namelen);
     }
 
+    if (!dot_found)			/* check for bad directories */
+    {
+	status("'.' directory missing ! (press any key)", 1, 1, 1, MSG_ERROR);
+	return 0;
+    }
+
+    if (!dotdot_found)			/* check for bad directories */
+    {
+	status("'..' directory missing ! (press any key)", 1, 1, 1, MSG_ERROR);
+	return 0;
+    }
+
     if (this->entries == MAX_ENTRIES && readdir(this->dir))
     {
-        sprintf(temp, "Too many directory entries!\
+	sprintf(temp, "Too many directory entries!\
  Only %d will be displayed. (press any key)", MAX_ENTRIES);
-        status(temp, 1, 1, 1, MSG_ERROR);
+	status(temp, 1, 1, 1, MSG_ERROR);
     }
-    
+
     if (verify)    
     {
         for (i = 0; i < old_entries; i++)
@@ -592,16 +719,18 @@ int panel_getdirinfo(panel *this, char *directory, int verify)
         free(old_dir_entry);
     }
 
+    CurrentSortMethod = this->sort_method;
+
     if (backdir_index != -1)
     {
         tmp = this->dir_entry[0];
         this->dir_entry[0] = this->dir_entry[backdir_index];
         this->dir_entry[backdir_index] = tmp;
-        qsort(this->dir_entry + 1, this->entries - 1,
+	qsort(this->dir_entry + 1, this->entries - 1,
 	      sizeof(_dir_entry), sortfn);
     }
     else
-        qsort(this->dir_entry, this->entries, sizeof(_dir_entry), sortfn);
+	qsort(this->dir_entry, this->entries, sizeof(_dir_entry), sortfn);
 
     return 1;
 }
@@ -646,7 +775,7 @@ void panel_update(panel *this)
     limit = min(this->lines - 2, this->on_screen);
     for (; i < limit; i++)
     {
-        window_cursormove(this->win, i - this->first_on_screen + 1, 1);
+        window_cursormove_notify(this->win, i - this->first_on_screen + 1, 1);
         window_write(temp, this->columns - 2);
     }
     this->on_screen = this->entries;
@@ -657,7 +786,7 @@ void panel_update(panel *this)
 void panel_trunc_fname(char *fname, char *dest, int len)
 {
     int flen;
-         
+
     if ((flen = strlen(fname)) > len)
     {
         dest[0] = dest[1] = dest[2] = '.';
@@ -674,33 +803,40 @@ void panel_update_path(panel *this)
     int i, len;
 
     memset(temp, ' ', this->columns);
-    panel_trunc_fname(this->path, t = temp, len = this->columns - 4 - 11);
+    panel_trunc_fname(this->path, t = temp, len = this->columns - 4 - 11 - 1);
+
     for (i = 0; i < len; i++)
-	if (t[i] < ' ' || t[i] > '~') t[i] = '?';
+	if (!is_print(t[i])) t[i] = '?';
 
     tty_bright(PanelPathBrightness);
     tty_foreground(PanelPath);
     tty_background(PanelFrame);
-    window_cursormove(this->win, 0, 2);
+    window_cursormove_notify(this->win, 0, 2);
     window_write(temp, len);
 }
 
 
 void panel_update_size(panel *this)
 {
-#ifdef HAVE_STATFS
-    struct statfs fsbuf;
     char sz[16];
 
+#ifdef HAVE_STATFS
+    struct statfs fsbuf;
+
     statfs(this->path, &fsbuf);
-    sprintf(sz, "%10d", fsbuf.f_bfree * fsbuf.f_bsize);
+    sprintf(sz, "%11d", fsbuf.f_bfree * fsbuf.f_bsize);
 
     tty_bright(PanelDeviceFreeSpaceBrightness);
     tty_foreground(PanelDeviceFreeSpace);
-    tty_background(PanelFrame);
-    window_cursormove(this->win, 0, this->columns - 2 - 10);
-    window_write(sz, strlen(sz));
+#else
+    memset(sz, ' ', 11);
+    tty_bright(OFF);
+    tty_foreground(PanelFrame);
 #endif
+
+    tty_background(PanelFrame);
+    window_cursormove_notify(this->win, 0, this->columns - 2 - 11);
+    window_write(sz, 11);
 }
 
 
@@ -716,21 +852,18 @@ void panel_update_info(panel *this)
             total_files ++;
             total_size += this->dir_entry[index].size;
         }
+
         sprintf(str, "%d bytes in %d file(s)", total_size, total_files);
         tty_bright(PanelFilesInfoBrightness);
         tty_foreground(PanelFilesInfo);
-        this->info_dirty = ON;
     }
     else
     {
 	if (InfoDisplay == OFF)
-	    if (this->info_dirty)
-	    {
-		*str = 0;
-		this->info_dirty = 0;
-	        goto display_info;
-	    }
-	    else return;
+	{
+	    *str = 0;
+	    goto display_info;
+	}
 
         strcpy(temp_rights, rights);
         mode = this->dir_entry[this->current_entry].mode;
@@ -759,7 +892,7 @@ void panel_update_info(panel *this)
 
 	memcpy(t = str, this->dir_entry[this->current_entry].name, len);
 	for (i = 0; i < len; i++)
-	    if (t[i] < ' ' || t[i] > '~') t[i] = '?';
+	    if (!is_print(t[i])) t[i] = '?';
 	memset(str + len, ' ', maxname - len);
 
         if (this->dir_entry[this->current_entry].type == DIR_ENTRY)
@@ -775,10 +908,11 @@ void panel_update_info(panel *this)
         tty_bright(PanelFileInfoBrightness);
         tty_foreground(PanelFileInfo);
     }    
+
     memcpy(temp, str, len = strlen(str));
     memset(temp + len, ' ', this->columns - 2 - len);
     tty_background(PanelFrame);
-    window_cursormove(this->win, this->lines - 1, 2);
+    window_cursormove_notify(this->win, this->lines - 1, 2);
     window_write(temp, this->columns - 4);
 }
 
@@ -790,12 +924,15 @@ void panel_update_entry(panel *this, int entry)
     int foreground, background, bright;
 
     memset(temp, ' ', this->columns);
-    reserved = (this->mode == 4) ? 4 : 20;
+    reserved = (this->display_mode == 4) ? 4 : 20;
     len = min(strlen(this->dir_entry[entry].name),
 	      (unsigned)this->columns - reserved);
+
     memcpy(t = &temp[1], this->dir_entry[entry].name, len);
+
     for (i = 0; i < len; i++)
-	if (t[i] < ' ' || t[i] > '~') t[i] = '?';
+	if (!is_print(t[i])) t[i] = '?';
+
     if (len == (unsigned)this->columns - reserved) len--;
     
     if (entry || this->path[1] == 0)
@@ -818,9 +955,9 @@ void panel_update_entry(panel *this, int entry)
                                 break;
         }
 
-    switch (this->mode)
+    switch (this->display_mode)
     {
-        case 0:
+        case DISPLAY_OWNER_GROUP:
 
             memcpy(temp + this->columns - 2 - 16,
 		   this->dir_entry[entry].owner, 7);
@@ -828,19 +965,19 @@ void panel_update_entry(panel *this, int entry)
 		   this->dir_entry[entry].group, 7);
 	    break;
 
-	case 1:
+	case DISPLAY_DATE_TIME:
 
 	    memcpy(temp + this->columns - 2 - 16,
 		   this->dir_entry[entry].date, 15);
 	    break;
 
-	case 2:
+	case DISPLAY_SIZE:
 
 	    sprintf(buf, "%10d", this->dir_entry[entry].size);
 	    memcpy(temp + this->columns - 2 - 11, buf, 10);
 	    break;
 
-	case 3:
+	case DISPLAY_MODE:
 
 	    strcpy(temp_rights, rights);
 
@@ -869,7 +1006,7 @@ void panel_update_entry(panel *this, int entry)
 	    memcpy(temp + this->columns - 2 - 11, temp_rights, 10);
 	    break;
 
-	case 4:
+	case DISPLAY_FULL_NAME:
 
 	    break;
 
@@ -877,7 +1014,7 @@ void panel_update_entry(panel *this, int entry)
 
 	    fatal("invalid mode");
     }
-    
+
     if (entry == this->current_entry && this->focus == ON)
     {
 	foreground = this->dir_entry[entry].selected ?
@@ -898,7 +1035,7 @@ void panel_update_entry(panel *this, int entry)
     tty_background(background);
     tty_bright(bright);
 
-    window_cursormove(this->win, entry - this->first_on_screen + 1, 1);
+    window_cursormove_notify(this->win, entry - this->first_on_screen + 1, 1);
     window_write(temp, this->columns - 2);
 }
 
@@ -907,36 +1044,39 @@ void panel_update_frame(panel *this)
 {
     int i;
     tty_status status;
-    char buf[256];
-    
 
     tty_save(&status);
-   
+
     tty_foreground(PanelFrame);
+    tty_background(PanelFrame);
     tty_bright(OFF);
-    tty_reverse(ON);
-    
+
     if (FrameDisplay == ON)
     {
-	for (i = 0; i < this->lines; i++)
+	for (i = 1; i < this->lines - 1; i++)
 	{
-	    window_cursormove(this->win, i, 0);
-	    tty_putch(' ');
+	    window_cursormove_notify(this->win, i, 0);
+	    window_putch(' ');
 	}
 
-	for (i = 0; i < this->lines; i++)
+	for (i = 1; i < this->lines - 1; i++)
 	{
-	    window_cursormove(this->win, i, this->columns - 1);
-	    tty_putch(' ');
+	    window_cursormove_notify(this->win, i, this->columns - 1);
+	    window_putch(' ');
 	}
     }
-    
-    memset(buf, ' ', 256);
-    window_cursormove(this->win, 0, 0);
-    tty_write(buf, this->columns);
 
-    window_cursormove(this->win, this->lines - 1, 0);
-    tty_write(buf, this->columns);
+    window_cursormove_notify(this->win, 0, 0);
+    window_write("  ", 2);
+    window_cursormove_notify(this->win, 0, this->columns - 11 - 1 - 2);
+    window_putch(' ');
+    window_cursormove_notify(this->win, 0, this->columns - 2);
+    window_write("  ", 2);
+
+    window_cursormove_notify(this->win, this->lines - 1, 0);
+    window_write("  ", 2);
+    window_cursormove_notify(this->win, this->lines - 1, this->columns - 2);
+    window_write("  ", 2);
 
     tty_restore(&status);
 }
@@ -1055,21 +1195,21 @@ int panel_copy(panel *this, char *src, char *dest, int mode)
         close(shandle);
         return D_CREATERR;
     }
-    
+
     memsize = min(len = filelength(shandle), COPY_BUFFER_SIZE);
     if (S_ISBLK(mode) || S_ISCHR(mode))
     {
         len = MAXINT;
         memsize = COPY_BUFFER_SIZE;
     }
-    
+
     if (len == 0)
     {
 	close(shandle);
 	close(dhandle);
 	return SD_OK;
     }
-    
+
     if ((buf = (char *)malloc(memsize)) == NULL)
     {
         close(shandle);
@@ -1146,7 +1286,7 @@ char moveerr[7][50] =
     " ",
     "can't create destination file !",
     "can't delete source file !",
-};    
+};
 
 
 int panel_move(panel *this, char *from, char *to)
@@ -1217,7 +1357,7 @@ int panel_verify_name(char *file_name)
             return 0;
         }
     return 1;
-}            		            
+}
 
 
 int panel_getindex(panel *this, char *str)
@@ -1244,12 +1384,12 @@ int panel_getindex(panel *this, char *str)
 int panel_action(panel *this, int action, panel *link,
 		 void *aux_info, int repeat_count)
 {
+    char oldpath[PATH_MAX], *ptr;
     int need_update, need_update_all, old_current_entry;
     int done = 0, i, err = 0, entry, len, back;
     int rename_dir = 0, verify = aux_info == (void *)-1;
     char oldentry[256], msg[PATH_MAX], text[PATH_MAX];
-    char oldpath[PATH_MAX], *ptr;
-    
+
     switch (action)
     {
         case act_ENTER:
@@ -1274,16 +1414,22 @@ int panel_action(panel *this, int action, panel *link,
 			    status(no_perm, 1, 1, 1, MSG_ERROR);
 			break; 
 		    }
+
 		    if (back)
 		    {
 			for (ptr = oldpath + strlen(oldpath) - 1;
 			     *ptr != '/';
 			     ptr--);
+
 			ptr++;
+
 			for (i = 0;
 			     strcmp(this->dir_entry[i].name, ptr) &&
 			     i < this->entries;
 			     i++);
+
+			if (i == this->entries) i = 0;
+
 			this->current_entry = i;
 			this->first_on_screen = max(0,
 						    this->current_entry -
@@ -1300,7 +1446,7 @@ int panel_action(panel *this, int action, panel *link,
 		    panel_update_size(link);
 		    done = 1;
 		    break;
-         			         
+
 		case FILE_ENTRY:
 
 		    if (this->dir_entry[this->current_entry].executable)
@@ -1308,7 +1454,7 @@ int panel_action(panel *this, int action, panel *link,
 			uitclock(OFF);
 			tty_end();
 			tty_putscreen((char *)aux_info);
-			sprintf(text, "\"%s\"",
+			sprintf(text, "exec ./\"%s\"",
 				this->dir_entry[this->current_entry].name);
 			restore_signals();
 			signals_dfl();
@@ -1317,6 +1463,7 @@ int panel_action(panel *this, int action, panel *link,
 			ignore_signals();
 			write(1, "\n\n", 2);
 			tty_init();
+			tty_touch();
 			panel_nooptimizations(this);
 			panel_nooptimizations(link);
 			edit_puts(this->edt,
@@ -1335,7 +1482,7 @@ int panel_action(panel *this, int action, panel *link,
 		signals(OFF);
 		if (this->dir_entry[this->current_entry].type == DIR_ENTRY)
 		    break;
-		sprintf(msg, "(COPY) Copying file %s to : ",
+		sprintf(msg, "(COPY) Copying file %s to: ",
 		       cutname(this->dir_entry[this->current_entry].name,0,0));
 		strcpy(oldpath, link->path);
 		strcat(oldpath, "/");
@@ -1361,7 +1508,7 @@ int panel_action(panel *this, int action, panel *link,
 		if (err != SD_OK)
 		{
 		    if (err == SD_CANCEL) break;
-		    sprintf(msg, "(COPY) Error copying file %s : %s",
+		    sprintf(msg, "(COPY) Error copying file %s: %s",
 			    cutname(this->dir_entry[this->current_entry].name,
 				    0, 1),
 			    copyerr[err]);
@@ -1374,7 +1521,7 @@ int panel_action(panel *this, int action, panel *link,
 		panel_update_size(link);
 	    }
 	    else
-	    {    
+	    {
 		signals(OFF);
 		if (strcmp(this->path, link->path) == 0)
 		{
@@ -1386,12 +1533,12 @@ int panel_action(panel *this, int action, panel *link,
 (ENTER to continue, TAB to cancel)",
 			   1, 0, 1, MSG_WARNING) != KEY_ENTER)
 		break;
-		
+
 		strcpy(temp, link->path);
 		strcat(temp, "/");
 		len = strlen(temp);
 		signals(ON);
-		
+
 		while ((entry = panel_getnext(this)) != -1)
 		{
 		    strcpy(temp + len, this->dir_entry[entry].name);
@@ -1402,7 +1549,7 @@ int panel_action(panel *this, int action, panel *link,
 			       1, 1, 1, MSG_ERROR);
 			break;
 		    }
-		    
+
 		    err = panel_copy(this, this->dir_entry[entry].name,
 				     temp, this->dir_entry[entry].mode);
 
@@ -1410,7 +1557,7 @@ int panel_action(panel *this, int action, panel *link,
 		    {
 			if (err == SD_CANCEL) break;
 
-			sprintf(msg, "(COPY) Error copying file %s : %s",
+			sprintf(msg, "(COPY) Error copying file %s: %s",
 				cutname(this->dir_entry[entry].name, 0, 1),
 				copyerr[err]);
 
@@ -1419,13 +1566,13 @@ int panel_action(panel *this, int action, panel *link,
 		    }
 		    else
 			this->dir_entry[entry].selected = 0;
-		    
+
 		    panel_update_size(this);
 		    panel_update_size(link);
 		}
 		status(NULL, 0, 0, 1, MSG_OK);
 	    }
-	    
+
 	    signals(OFF);
 
 	    if (!panel_getdirinfo(link, link->path, ON))
@@ -1444,7 +1591,7 @@ int panel_action(panel *this, int action, panel *link,
 		panel_update_info(this);
 	    }
 	    break;
-	    
+
         case act_DELETE:
 
 	    if (!this->selected_files &&
@@ -1486,7 +1633,7 @@ int panel_action(panel *this, int action, panel *link,
 		panel_update_size(link);
 	    }
 	    else
-	    {    
+	    {
 		signals(OFF);
 
 		if (status("(DEL) Deleting file(s) ...  \
@@ -1498,7 +1645,7 @@ int panel_action(panel *this, int action, panel *link,
 		    if (this->dir_entry[i].selected) break;
 
 		signals(ON);
-		
+
 		while ((entry = panel_getnext(this)) != -1)
 		{
 		    sprintf(msg, "(DEL) Deleting %s",
@@ -1534,7 +1681,7 @@ int panel_action(panel *this, int action, panel *link,
 		panel_update_size(link);
 		status(NULL, 0, 0, 1, MSG_OK);
 	    }
-	    
+
 	    signals(OFF);
 
 	    if (!panel_getdirinfo(this, this->path, ON))
@@ -1576,8 +1723,8 @@ int panel_action(panel *this, int action, panel *link,
 
 	    panel_action(this, act_DOWN, link, NULL, repeat_count);
 	    break;
-	    
-        case act_SELECTALL:
+
+        case act_SELECT_ALL:
 
 	    for (i = 0; i < this->entries; i++)
 	        if (this->dir_entry[i].type != DIR_ENTRY)
@@ -1590,7 +1737,7 @@ int panel_action(panel *this, int action, panel *link,
 	    done = 1;
 	    break;
 
-        case act_UNSELECTALL:
+        case act_UNSELECT_ALL:
 
 	    for (i = 0; i < this->entries; i++)
 	        this->dir_entry[i].selected = OFF;
@@ -1599,7 +1746,7 @@ int panel_action(panel *this, int action, panel *link,
 	    panel_update(this);
 	    done = 1;
 	    break;
-                       		
+
         case act_TOGGLE:
 
 	    for (i = 0; i < this->entries; i++)
@@ -1610,10 +1757,10 @@ int panel_action(panel *this, int action, panel *link,
 	    panel_update(this);
 	    done = 1;
 	    break;
-				
+
         case act_MKDIR:
 
-	    if (!edit_gets(this->edt, "(MKDIR) Enter directory name : ",
+	    if (!edit_gets(this->edt, "(MKDIR) Enter directory name: ",
                            temp, NULL))
 	        break;
 
@@ -1650,7 +1797,7 @@ int panel_action(panel *this, int action, panel *link,
 
 	    panel_update_size(link);
 	    break;
-	    
+
         case act_MOVE:
 
             this->chkdest = ON;
@@ -1659,7 +1806,7 @@ int panel_action(panel *this, int action, panel *link,
                 (this->dir_entry[this->current_entry].type != DIR_ENTRY))
 	    {
 		signals(OFF);
-		sprintf(msg, "(MOVE) Moving file %s to : ",
+		sprintf(msg, "(MOVE) Moving file %s to: ",
                        cutname(this->dir_entry[this->current_entry].name,0,0));
 
 		strcpy(oldpath, link->path);
@@ -1680,24 +1827,24 @@ int panel_action(panel *this, int action, panel *link,
 		    status(msg, 1, 1, 1, MSG_ERROR);
 		    break;
 		}
-		
+
 		if ((err=panel_move(this,
                                     this->dir_entry[this->current_entry].name,
                                     text)) != FT_OK)
 		{
 		    if (err == FT_CANCEL) break;
-		    sprintf(msg, "(MOVE) Error moving file %s : %s",
+		    sprintf(msg, "(MOVE) Error moving file %s: %s",
                             cutname(this->dir_entry[this->current_entry].name,
                                     0, 1), moveerr[err]);
 		    status(msg, 1, 1, 1, MSG_ERROR);
 		}
-		
+
 		status(NULL, 0, 0, 1, MSG_OK);
 		panel_update_size(this);
 		panel_update_size(link);
 	    }
 	    else
-	    {    
+	    {
 		if (!this->selected_files &&
                     this->dir_entry[this->current_entry].type == DIR_ENTRY)
 		{
@@ -1706,14 +1853,14 @@ int panel_action(panel *this, int action, panel *link,
 		    if (strcmp(this->dir_entry[this->current_entry].name, "..")
                         == 0) break;
 
-		    if (!edit_gets(this->edt, "(RENAME) New directory name : ",
+		    if (!edit_gets(this->edt, "(RENAME) New directory name: ",
                                    text, NULL) || !text[0])
 		        break;
 
 		    if (rename(this->dir_entry[this->current_entry].name, text)
                         == -1)
 		    {
-			sprintf(msg, "(RENAME) Error renaming directory : %s !\
+			sprintf(msg, "(RENAME) Error renaming directory: %s !\
   (press any key)", cutname(this->dir_entry[this->current_entry].name, 0, 0));
 			status(msg, 1, 1, 1, MSG_ERROR);
 			break;
@@ -1740,7 +1887,7 @@ int panel_action(panel *this, int action, panel *link,
 		    strcat(temp, "/");
 		    len = strlen(temp);
 		    signals(ON);
-		    
+
 		    while ((entry = panel_getnext(this)) != -1)
 		    {
 			strcpy(temp + len, this->dir_entry[entry].name);
@@ -1757,7 +1904,7 @@ int panel_action(panel *this, int action, panel *link,
 			{
 			    if (err == FT_CANCEL) break;
 
-			    sprintf(msg, "(MOVE) Error moving file %s : %s",
+			    sprintf(msg, "(MOVE) Error moving file %s: %s",
                                     cutname(this->dir_entry[entry].name, 0, 1),
                                     moveerr[err]);
 
@@ -1815,7 +1962,7 @@ int panel_action(panel *this, int action, panel *link,
 		    need_update_all = 1;
 		}
 		else
-		{ 
+		{
 		    this->current_entry--;
 		    if (!need_update)
 		        panel_update_entry(this, this->current_entry + 1);
@@ -1856,7 +2003,7 @@ int panel_action(panel *this, int action, panel *link,
 	        else
 	            done = -1;
 	    break;
-	    
+
 	case act_PGUP:
 
 	    if (this->current_entry == 0)
@@ -1864,7 +2011,7 @@ int panel_action(panel *this, int action, panel *link,
 		done = -1;
 		break;
 	    }
-	    
+
 	    old_current_entry = this->current_entry;
 	    
 	    if (this->current_entry < this->lines - 2)
@@ -1875,7 +2022,7 @@ int panel_action(panel *this, int action, panel *link,
 		this->first_on_screen = max(0, this->first_on_screen -
                                                (this->lines - 2));
 	    }
-	    
+
 	    if (this->entries > this->lines - 2)
 	        panel_update(this);
 	    else
@@ -1883,7 +2030,7 @@ int panel_action(panel *this, int action, panel *link,
 		panel_update_entry(this, old_current_entry);
 		panel_update_entry(this, this->current_entry);
 	    }
-	    
+
 	    break;
 
         case act_PGDOWN:
@@ -1893,12 +2040,12 @@ int panel_action(panel *this, int action, panel *link,
 		done = -1;
 		break;
 	    }
-	    
+
 	    old_current_entry = this->current_entry;
 	    
 	    if (this->entries - 1 - this->first_on_screen < this->lines - 2)
 	        this->current_entry = this->entries - 1;
-	    else   
+	    else
 	        if (this->entries - 1 - this->current_entry < this->lines - 2)
 	        {
 		    this->current_entry = this->entries - 1;
@@ -1912,7 +2059,7 @@ int panel_action(panel *this, int action, panel *link,
                         min(this->first_on_screen + this->lines - 2,
                         (this->entries - 1) - (this->lines - 2) + 1);
 	         }
-	    
+
 	    if (this->entries > this->lines - 2)
 	        panel_update(this);
 	    else
@@ -1920,7 +2067,7 @@ int panel_action(panel *this, int action, panel *link,
 		panel_update_entry(this, old_current_entry);
 		panel_update_entry(this, this->current_entry);
 	    }
-	    
+
 	    break;
 
         case act_HOME:
@@ -1936,7 +2083,7 @@ int panel_action(panel *this, int action, panel *link,
 	    this->current_entry = this->entries - 1;
 	    this->first_on_screen = max(0, (this->entries - 1) -
                                            (this->lines - 2) + 1);
-	    panel_update(this);        
+	    panel_update(this);
 	    break;
 
 	case act_CHDIR:
@@ -1944,11 +2091,56 @@ int panel_action(panel *this, int action, panel *link,
 	    this->first_on_screen = this->current_entry = 0;
 	    strcpy(this->path, (char *)aux_info);
 	    panel_action(this, act_REFRESH, NULL, NULL, 1);
+	    if (strcmp(this->path, link->path) == 0)
+		panel_action(link, act_REFRESH, NULL, NULL, 1);
 	    break;
 
-	case act_MODE:
+	case act_DISPLAY_NEXT_MODE:
+
+	    this->display_mode = (this->display_mode + 1) % FILE_DISPLAY_MODES;
+	    goto all_display_modes;
+
+	case act_DISPLAY_OWNER_GROUP:
+	case act_DISPLAY_DATE_TIME:
+	case act_DISPLAY_SIZE:
+	case act_DISPLAY_MODE:
+	case act_DISPLAY_FULL_NAME:
+
+	    this->display_mode = action - act_DISPLAY_OWNER_GROUP;
+
+	  all_display_modes:
+
+	    panel_update(this);
+	    break;
+
+	case act_SORT_NEXT_METHOD:
+
+	    this->sort_method = (this->sort_method + 1) % FILE_SORT_METHODS;
+	    goto all_sort_methodes;
+
+	case act_SORT_BY_NAME:
+	case act_SORT_BY_EXTENSION:
+	case act_SORT_BY_SIZE:
+	case act_SORT_BY_DATE:
+	case act_SORT_BY_MODE:
+	case act_SORT_BY_OWNER_ID:
+	case act_SORT_BY_GROUP_ID:
+	case act_SORT_BY_OWNER_NAME:
+	case act_SORT_BY_GROUP_NAME:
+
+	    this->sort_method = action - act_SORT_BY_NAME;
+
+	  all_sort_methodes:
+
+	    CurrentSortMethod = this->sort_method;
+
+	    if (this->path[0] == 0) /* root dir */
+		qsort(this->dir_entry, this->entries,
+		      sizeof(_dir_entry), sortfn);
+	    else
+		qsort(this->dir_entry + 1, this->entries - 1,
+		      sizeof(_dir_entry), sortfn);
 
-	    this->mode = (this->mode + 1) % FILE_DISPLAY_MODES;
 	    panel_update(this);
 	    break;
 
@@ -1966,6 +2158,46 @@ int panel_action(panel *this, int action, panel *link,
                                     link->lines, link->columns);
 	    break;
 
+	case act_PATTERN_SELECT:
+	case act_PATTERN_UNSELECT:
+	
+	    for (i = 0; i < this->entries; i++)
+		if (this->dir_entry[i].type != DIR_ENTRY)
+		{
+		    int fnm_flags = FNM_PATHNAME;
+		    if (LeadingDotMatch == OFF)
+		        fnm_flags |= FNM_PERIOD;
+
+		    if (fnmatch(aux_info,this->dir_entry[i].name,fnm_flags) !=
+		    	FNM_NOMATCH)
+		    {
+		        switch (action)
+		        {
+			    case act_PATTERN_SELECT:
+
+				if (!this->dir_entry[i].selected)
+				{
+				    this->dir_entry[i].selected = ON;
+				    this->selected_files++;
+				}
+				break;
+
+			    case act_PATTERN_UNSELECT:
+
+				if (this->dir_entry[i].selected)
+				{
+				    this->dir_entry[i].selected = OFF;
+				    this->selected_files--;
+				}
+				break;
+			}
+		    }
+		}
+
+	    panel_update(this);
+	    done = 1;
+	    break;
+
 	case act_REFRESH:
 
 	    if (this->dir_entry[this->current_entry].name)
