@@ -1,6 +1,6 @@
 /*
     Status line management header file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -18,8 +18,8 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 
-#ifndef __STATUS_H
-#define __STATUS_H
+#ifndef _STATUS_H
+#define _STATUS_H
 
 
 #include "config.h"
@@ -36,4 +36,4 @@ void status_end(void);
 int  status(char *msg_name, int wait, int sound, int restore, int msg_type);
 
 
-#endif
+#endif	/* _STATUS_H */
