@@ -1,6 +1,6 @@
 /*
     Terminal capabilities interface header file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -18,8 +18,8 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 
-#ifndef __TERMCAP_H
-#define __TERMCAP_H
+#ifndef _TERMCAP_H
+#define _TERMCAP_H
 
 
 #define TTY_DESCRIPTION_END	 0
@@ -40,4 +40,4 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 void termcap_init(void);
 
 
-#endif
+#endif	/* _TERMCAP_H */
