@@ -1,6 +1,6 @@
 /*
     UIT main file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -57,7 +57,7 @@ int  suspended = OFF;
 int  suspend_allowed = OFF;
 char cSection[]  = "[UIT-Color]";
 char bwSection[] = "[UIT-Monochrome]";
-char title[] = "UNIX Interactive Tools 4.2b";
+char title[] = " UNIX Interactive Tools 4.3";
 char login[] = "User:";
 char tty[]   = "tty:";
 char exit_msg[] = "*** (EXIT) Exit UNIX Interactive Tools ? \
@@ -73,7 +73,8 @@ extern int FrameDisplay;
 extern char tty_type[];
 static int UserHeartAttack;
 
-#define BUILTIN_OPERATIONS		25
+
+#define BUILTIN_OPERATIONS		47
 
 #define BUILTIN_Copy			-1
 #define BUILTIN_RenameMove		-2
@@ -86,23 +87,45 @@ static int UserHeartAttack;
 #define BUILTIN_GotoRoot		-9
 #define BUILTIN_SwitchPanels		-10
 #define BUILTIN_HistoryNext		-11
-#define BUILTIN_FileDisplayMode		-12
-#define BUILTIN_DeleteCmdLn		-13
-#define BUILTIN_FileSelect		-14
-#define BUILTIN_CopyFileToCmdLn		-15
-#define BUILTIN_Home			-16
-#define BUILTIN_End			-17
-#define BUILTIN_PageUp			-18
-#define BUILTIN_PageDown		-19
-#define BUILTIN_CursorLeft		-20
-#define BUILTIN_CursorRight		-21
-#define BUILTIN_CursorUp		-22
-#define BUILTIN_CursorDown		-23
-#define BUILTIN_ChangePanel		-24
-#define BUILTIN_ChangeDir		-25
+#define BUILTIN_FileDisplayNextMode	-12
+#define BUILTIN_FileDisplayOwnerGroup	-13
+#define BUILTIN_FileDisplayDateTime	-14
+#define BUILTIN_FileDisplaySize		-15
+#define BUILTIN_FileDisplayMode		-16
+#define BUILTIN_FileDisplayFullName	-17
+#define BUILTIN_FileSortNextMethod	-18
+#define BUILTIN_FileSortByName		-19
+#define BUILTIN_FileSortByExtension	-20
+#define BUILTIN_FileSortBySize		-21
+#define BUILTIN_FileSortByDate		-22
+#define BUILTIN_FileSortByMode		-23
+#define BUILTIN_FileSortByOwnerId	-24
+#define BUILTIN_FileSortByGroupId	-25
+#define BUILTIN_FileSortByOwnerName	-26
+#define BUILTIN_FileSortByGroupName	-27
+#define BUILTIN_DeleteCmdLn		-28
+#define BUILTIN_FileSelect		-29
+#define BUILTIN_CopyFileToCmdLn		-30
+#define BUILTIN_Home			-31
+#define BUILTIN_End			-32
+#define BUILTIN_PageUp			-33
+#define BUILTIN_PageDown		-34
+#define BUILTIN_CursorLeft		-35
+#define BUILTIN_CursorRight		-36
+#define BUILTIN_CursorUp		-37
+#define BUILTIN_CursorDown		-38
+#define BUILTIN_ChangePanel		-39
+#define BUILTIN_ChangeDir		-40
+#define BUILTIN_HardRefresh		-41
+#define BUILTIN_FilePatternSelect	-42
+#define BUILTIN_FilePatternUnselect	-43
+#define BUILTIN_DeleteCmdLnWord		-44
+#define BUILTIN_ConformCurrentDir	-45
+#define BUILTIN_ConformOtherDir		-46
+#define BUILTIN_CopyPathToCmdLn		-47
 
 
-#define MAX_BUILTIN_NAME		20
+#define MAX_BUILTIN_NAME		25
 
 
 char built_in[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
@@ -118,7 +141,22 @@ char built_in[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
     "<GotoRoot>",
     "<SwitchPanels>",
     "<HistoryNext>",
+    "<FileDisplayNextMode>",
+    "<FileDisplayOwnerGroup>",
+    "<FileDisplayDateTime>",
+    "<FileDisplaySize>",
     "<FileDisplayMode>",
+    "<FileDisplayFullName>",
+    "<FileSortNextMethod>",
+    "<FileSortByName>",
+    "<FileSortByExtension>",
+    "<FileSortBySize>",
+    "<FileSortByDate>",
+    "<FileSortByMode>",
+    "<FileSortByOwnerId>",
+    "<FileSortByGroupId>",
+    "<FileSortByOwnerName>",	
+    "<FileSortByGroupName>",
     "<DeleteCmdLn>",
     "<FileSelect>",
     "<CopyFileToCmdLn>",
@@ -132,6 +170,13 @@ char built_in[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
     "<CursorDown>",
     "<ChangePanel>",
     "<ChangeDir>",
+    "<HardRefresh>",
+    "<FilePatternSelect>",
+    "<FilePatternUnselect>",
+    "<DeleteCmdLnWord>",
+    "<ConformCurrentDir>",
+    "<ConformOtherDir>",
+    "<CopyPathToCmdLn>",
 };
 
 
@@ -242,7 +287,9 @@ void fatal_signal(int signum)
 	    fprintf(stderr, "Please report to tudor@ulise.cs.pub.ro\n");
 	    break;
     }
+
     removelog();
+    tty_exit();
     exit(1);
 }
 
@@ -271,8 +318,6 @@ void ignore_signals(void)
     signal(SIGABRT, SIG_IGN);
     signal(SIGUSR1, SIG_IGN);
     signal(SIGUSR2, SIG_IGN);
-    signal(SIGTTIN, SIG_IGN);
-    signal(SIGTTOU, SIG_IGN);
     signal(SIGPIPE, SIG_IGN);
     signal(SIGFPE,  SIG_IGN);
 }
@@ -285,8 +330,6 @@ void restore_signals(void)
     signal(SIGABRT, SIG_DFL);
     signal(SIGUSR1, SIG_DFL);
     signal(SIGUSR2, SIG_DFL);
-    signal(SIGTTIN, SIG_DFL);
-    signal(SIGTTOU, SIG_DFL);
     signal(SIGPIPE, SIG_DFL);
     signal(SIGFPE,  SIG_DFL);
 }
@@ -294,8 +337,9 @@ void restore_signals(void)
 
 void settitle(void)
 {
-    char buf[256], *login_name = getlogin(), *tty_name = ttyname(1);
+    int len;
     window *title_win = window_init(0, 0, 1, SCREEN_X);
+    char buf[256], *login_name = getlogin(), *tty_name = ttyname(1);
 
     if (login_name == NULL) login_name = "";
     if (tty_name   == NULL) tty_name   = "";
@@ -303,14 +347,17 @@ void settitle(void)
     tty_bright(TitleBrightness);
     tty_foreground(TitleForeground);
     tty_background(TitleBackground);
-    window_cursormove(title_win, 0, 0);
-    memset(buf, ' ', SCREEN_X);
-    window_write(buf, SCREEN_X);
 
-    sprintf(buf, "%s %s  %s %s        ", login, login_name, tty, tty_name);
-    window_cursormove(title_win, 0, 1);
+    window_cursormove_notify(title_win, 0, 0);
     window_write(title, strlen(title));
-    window_cursormove(title_win, 0, SCREEN_X - strlen(buf)  - 1);
+
+    len = strlen(login) + 1 + strlen(login_name) + 2 + strlen(tty) + 1 + strlen(tty_name);
+
+    memset(buf, ' ', len = SCREEN_X - strlen(title) - len - 2 - 6 - 1);
+    window_cursormove_notify(title_win, 0, strlen(title));
+    window_write(buf, len);
+
+    window_cursormove_notify(title_win, 0, strlen(title) + len);
     window_write(login, strlen(login));
     window_putch(' ');
     tty_foreground(UserName);
@@ -322,6 +369,12 @@ void settitle(void)
     window_putch(' ');
     tty_foreground(TtyName);
     window_write(tty_name, strlen(tty_name));
+
+    tty_foreground(TitleForeground);
+    window_putch(' ');
+    window_putch(' ');
+    window_cursormove_notify(title_win, 0, SCREEN_X - 1);
+    window_putch(' ');
 }
 
 
@@ -341,7 +394,7 @@ void clock_on(int signum)
     if ((hour = _tm->tm_hour % 12) == 0) hour = 12;
     sprintf(buf, "%2d:%02d%c",
 	    hour, _tm->tm_min, (_tm->tm_hour < 12) ? 'a' : 'p');
-    tty_cursormove(0, SCREEN_X - 7);
+    tty_cursormove_notify(0, SCREEN_X - 7);
     tty_foreground(ClockForeground);
     tty_background(ClockBackground);
     tty_bright(ClockBrightness);
@@ -409,8 +462,9 @@ int start(char *cmd, int hide)
 	stderr_log = fopen(fname, "w");
 	
 	restore_signals();
+ 	sprintf(fname, "exec %s", cmd);
 	signals_dfl();
-        child_exit_code = system(cmd);
+        child_exit_code = system(fname);
 	signals(SignalsStatus);
 	ignore_signals();
 	
@@ -426,8 +480,9 @@ int start(char *cmd, int hide)
         tty_putscreen(screen);
 
 	restore_signals();
+ 	sprintf(fname, "exec %s", cmd);
 	signals_dfl();
-        child_exit_code = system(cmd);
+        child_exit_code = system(fname);
         signals(SignalsStatus);
 	ignore_signals();
 
@@ -443,8 +498,9 @@ void fatal(char *postmsg)
     uitclock(OFF);
     tty_end();
     tty_clrscr();
-    fprintf(stderr, "Fatal error : %s !\n", postmsg);
+    fprintf(stderr, "Fatal error: %s !\n", postmsg);
     removelog();
+    tty_exit();
     exit(1);
 }
 
@@ -693,8 +749,8 @@ int main(void)
     signals(OFF);
     ignore_signals();
 
-    printf("UNIX Interactive Tools 4.2b\n");
-    printf("Copyright (c) May 1994 by Tudor Hulubei & Andrei Pitis\n");
+    printf("UNIX Interactive Tools 4.3\n");
+    printf("Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis\n");
 
     termcap_init();
     tty_kbdinit(0);
@@ -844,6 +900,7 @@ int main(void)
 	      insert:
 
                 ConfigKeyInfo->OpBuiltin = 0;
+
 	        for (j = 0; j < BUILTIN_OPERATIONS; j++)
 	            if (strcmp(ConfigKeyInfo->OpName, built_in[j]) == 0)
 	            {
@@ -852,6 +909,7 @@ int main(void)
 	                ConfigKeyInfo->OpBuiltin = 1;
 	                break;
 	            }
+
 		if (ConfigKeyInfo->OpBuiltin || ConfigKeyInfo->OpCommand)
 		    if (tty_key_convert(key_seq))
 		        tty_key_list_insert(key_seq, (void *)ConfigKeyInfo);
@@ -874,6 +932,7 @@ int main(void)
     }
 
     tty_getsize(&SCREEN_X, &SCREEN_Y);
+    tty_startup();
 
 #ifdef HAVE_LINUX
     if (LinuxConsole)
@@ -962,6 +1021,7 @@ restart:
 	    suspended = OFF;
 	    panel_nooptimizations(current_panel);
 	    panel_nooptimizations(other_panel);
+	    tty_touch();
             goto restart;
 	}
 
@@ -1000,18 +1060,17 @@ restart:
 				   status("*** Can't open stderr log file ***",
 					  1, 1, 1, MSG_ERROR);
 				else
-				{
 				    while (fgets(cmdln, SCREEN_X, stderr_log))
 				        if (status(cmdln,1,0,0,MSG_ERROR) !=
 					    KEY_ENTER)
 					    break;
-				}
 				fclose(stderr_log);
 			    }
 			    else
 			    {
 				if (ConfigKeyInfo->OpSaveScreen)
 				    tty_getscreen(screen);
+				tty_touch();
 				if (FrameDisplay == OFF) need_clrscr = 1;
 				if (ConfigKeyInfo->OpPause) wait = 1;
 			    }
@@ -1054,6 +1113,9 @@ restart:
 		other_panel = temp_panel;
 		panel_no = !panel_no;
 		panel_setfocus(current_panel, ON, other_panel);
+
+	    restore:
+
 		edit_gettext(edt, cmdln);
 		edit_reset(edt);
 		edit_puts(edt,
@@ -1093,14 +1155,14 @@ restart:
 			case '+':
 
 		            action_status = panel_action(current_panel,
-							 act_SELECTALL,
+							 act_SELECT_ALL,
 							 other_panel, NULL, 1);
 			    break;
 
 		        case '-':
 
 			    action_status = panel_action(current_panel,
-							 act_UNSELECTALL,
+							 act_UNSELECT_ALL,
 							 other_panel, NULL, 1);
 			    break;
 
@@ -1113,9 +1175,10 @@ restart:
 
 			default :
 
+			    start(cmdln, 0);
 		            panel_nooptimizations(current_panel);
 			    panel_nooptimizations(other_panel);
-			    start(cmdln, 0);
+			    tty_touch();
 			    action_status = -1; 
 			    break;
 		    }
@@ -1208,17 +1271,12 @@ restart:
 		signals(OFF);
 		uitclock_status(OFF);
 		panel_action(current_panel, act_CHDIR, other_panel, "/", 1);
-		edit_gettext(edt, cmdln);
-		edit_reset(edt);
-		edit_puts(edt, strcat(panel_getpath(current_panel,
-						    temppath[0],
-						    MAX_STATIC_SIZE),
-				      PS1));
-		edit_eos(edt);
-		edit_puts(edt, cmdln);
-		edit_update(edt);
-		uitclock_status(ON);
-		break;
+
+		goto restore;
+
+	    case BUILTIN_HardRefresh:
+	    
+	        tty_touch();
 
             case BUILTIN_Refresh:
 
@@ -1302,6 +1360,11 @@ restart:
 			    edit_del(edt);
 			    break;
 
+			case BUILTIN_DeleteCmdLnWord:
+
+			    edit_delete_last_word(edt);
+			    break;
+
 			case BUILTIN_Refresh:
 
 			    panel_nooptimizations(current_panel);
@@ -1334,6 +1397,7 @@ restart:
 
 		panel_nooptimizations(current_panel);
 		panel_nooptimizations(other_panel);
+		tty_touch();
 		status(NULL, 0, 0, 1, MSG_OK);
 
 		if (app_end)
@@ -1352,6 +1416,14 @@ restart:
 		uitclock_status(ON);
 		break;
 
+	    case BUILTIN_DeleteCmdLnWord:
+
+		signals(OFF);
+		uitclock_status(OFF);
+		edit_delete_last_word(edt);
+		uitclock_status(ON);
+		break;
+
             case BUILTIN_Copy:
 
 		uitclock_status(OFF);
@@ -1381,14 +1453,41 @@ restart:
 		uitclock_status(ON);
 		break;
 
+	    case BUILTIN_FileDisplayNextMode:
+
+	    case BUILTIN_FileDisplayOwnerGroup:
+	    case BUILTIN_FileDisplayDateTime:
+	    case BUILTIN_FileDisplaySize:
 	    case BUILTIN_FileDisplayMode:
+	    case BUILTIN_FileDisplayFullName:
 
 		signals(OFF);
 		uitclock_status(OFF);
+		panel_action(current_panel,
+			     act_DISPLAY_NEXT_MODE -
+			     (key - BUILTIN_FileDisplayNextMode),
+			     NULL, NULL, 1);
+		uitclock_status(ON);
+		break;
 
-		for (i = 0; i < repeat_count; i++)
-		    panel_action(current_panel, act_MODE, NULL, NULL, 1);
+	    case BUILTIN_FileSortNextMethod:
 
+	    case BUILTIN_FileSortByName:
+	    case BUILTIN_FileSortByExtension:
+	    case BUILTIN_FileSortBySize:
+	    case BUILTIN_FileSortByDate:
+	    case BUILTIN_FileSortByMode:
+	    case BUILTIN_FileSortByOwnerId:
+	    case BUILTIN_FileSortByGroupId:
+	    case BUILTIN_FileSortByOwnerName:
+	    case BUILTIN_FileSortByGroupName:
+
+		signals(OFF);
+		uitclock_status(OFF);
+		panel_action(current_panel,
+			     act_SORT_NEXT_METHOD -
+			     (key - BUILTIN_FileSortNextMethod),
+			     NULL, NULL, 1);
 		uitclock_status(ON);
 		break;
 
@@ -1433,6 +1532,37 @@ restart:
 		uitclock_status(ON);
 		break;
 
+            case BUILTIN_CopyPathToCmdLn:
+
+		signals(OFF);
+		uitclock_status(OFF);
+		edit_gettext(edt, cmdln);
+		len = strlen(cmdln);
+
+		if (len && isalnum(cmdln[len - 1]))
+		{
+		    uitclock_status(ON);
+		    break;
+		}
+
+		panel_getpath(other_panel, &cmdln[len], PATH_MAX);
+
+		while (cmdln[len])
+		{
+		    if (!is_print(cmdln[len]))
+		    {
+			tty_beep();
+			cmdln[len] = 0;
+			break;
+		    }
+		    else
+		        edit_putch(edt, cmdln[len++]);
+		}
+
+		edit_putch(edt, ' ');
+		uitclock_status(ON);
+		break;
+
 	    case BUILTIN_HistoryBack:
 
 		signals(OFF);
@@ -1474,19 +1604,95 @@ restart:
 
 		signals(OFF);
 		uitclock_status(OFF);
-		edit_gets(edt, "(CHDIR) Change to directory: ", cmdname, NULL);
-		panel_action(current_panel, act_CHDIR, other_panel, cmdname,1);
-		edit_gettext(edt, cmdln);
-		edit_reset(edt);
-		edit_puts(edt, strcat(panel_getpath(current_panel,
-						    temppath[0],
-						    MAX_STATIC_SIZE),
-				      PS1));
-		edit_eos(edt);
-		edit_puts(edt, cmdln);
-		edit_update(edt);
-		uitclock_status(ON);
-		break;
+
+		if (edit_gets(edt, "(CHDIR) Change to directory: ",
+			      cmdname, NULL))
+		{
+		    char tmp_path[PATH_MAX];
+
+		    if (cmdname[0] == 0) break;
+
+		    if (cmdname[0] == '~')
+		    {
+		        strcpy(tmp_path, getenv("HOME"));
+		        strcat(tmp_path, "/");
+		        strcat(tmp_path, cmdname + 1);
+		        panel_action(current_panel, act_CHDIR,
+		    		     other_panel, tmp_path, 1);
+		    }
+		    else
+		        panel_action(current_panel, act_CHDIR,
+		    		     other_panel, cmdname, 1);
+		}
+
+		goto restore;
+
+
+	    case BUILTIN_FilePatternSelect:
+
+		signals(OFF);
+		uitclock_status(OFF);
+
+		if (edit_gets(edt, "(SELECT) Pattern: ",
+			      cmdname, NULL))
+		{
+		    if (cmdname[0] == 0) break;
+
+		    panel_action(current_panel, act_PATTERN_SELECT,
+		    		 other_panel, cmdname, 1);
+		}
+
+		goto restore;
+
+	    case BUILTIN_FilePatternUnselect:
+
+		signals(OFF);
+		uitclock_status(OFF);
+
+		if (edit_gets(edt, "(UNSELECT) Pattern: ",
+			      cmdname, NULL))
+		{
+		    if (cmdname[0] == 0) break;
+
+		    panel_action(current_panel, act_PATTERN_UNSELECT,
+		    		 other_panel, cmdname, 1);
+		}
+
+		goto restore;
+
+	    case BUILTIN_ConformCurrentDir:
+
+		signals(OFF);
+		uitclock_status(OFF);
+
+		{
+		    char tmp_path[PATH_MAX];
+
+		    panel_action(current_panel, act_CHDIR, other_panel,
+				 panel_getpath(other_panel,
+				 	       tmp_path,
+				 	       PATH_MAX),
+				 1);
+		}
+
+		goto restore;
+
+	    case BUILTIN_ConformOtherDir:
+
+		signals(OFF);
+		uitclock_status(OFF);
+
+		{
+		    char tmp_path[PATH_MAX];
+
+		    panel_action(other_panel, act_CHDIR, current_panel,
+				 panel_getpath(current_panel,
+				 	       tmp_path,
+				 	       PATH_MAX),
+				 1);
+		}
+
+		goto restore;
 
             default:
 
@@ -1496,8 +1702,10 @@ restart:
 		    edit_putch(edt, key);
 		    uitclock_status(ON);
 		}
+
 		break;
 	}
+
 	signals(ON);
         edit_setcursor(edt);
     }
@@ -1527,5 +1735,6 @@ restart:
     tty_clrscr();
 #endif
 
+    tty_exit();
     return 0;
 }
