@@ -1,6 +1,6 @@
 /*
     Process viewer/killer utility
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -116,14 +116,14 @@ char *ps_vect[MAX_PROCESS];
 int first_on_screen, current_process;
 window *title_win, *header_win, *screen_win, *status_win;
 static char title_text[]  =
-    " UNIX Interactive Tools 4.2b - Process Viewer/Killer";
+    " UNIX Interactive Tools 4.3 - Process Viewer/Killer";
 static char UitPsModeHelp[256];
 static char no_perm[]     = "not owner !";
 static char no_proc[]     = "no such process ! (REFRESH recommended)";
 
 struct SIGNAL
 {
-    char signame[16];
+    char signame[8];
     int  signal;
 };
 
@@ -150,7 +150,7 @@ static struct SIGNAL signals[] =
 };
 
 
-#define BUILTIN_OPERATIONS	24
+#define BUILTIN_OPERATIONS	25
 
 
 #define BUILTIN_CursorUp	0
@@ -177,6 +177,7 @@ static struct SIGNAL signals[] =
 #define BUILTIN_KillProcess	21
 #define BUILTIN_Refresh		22
 #define BUILTIN_Exit		23
+#define BUILTIN_HardRefresh	24
 
 
 #define MAX_BUILTIN_NAME	15
@@ -206,7 +207,8 @@ char built_in[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
     "<SIGCONT>",
     "<KillProcess>",
     "<Refresh>",
-    "<Exit>"
+    "<Exit>",
+    "<HardRefresh>",
 };
 
 
@@ -232,7 +234,7 @@ void settitle(void)
     tty_foreground(TitleForeground);
     tty_background(TitleBackground);
 
-    window_cursormove(title_win, 0, 0);
+    window_cursormove_notify(title_win, 0, 0);
     window_write(temp, SCREEN_X);
 }
 
@@ -248,7 +250,7 @@ void setheader(void)
     tty_foreground(HeaderForeground);
     tty_background(HeaderBackground);
 
-    window_cursormove(header_win, 0, 0);
+    window_cursormove_notify(header_win, 0, 0);
     window_write(temp, SCREEN_X);
 }
 
@@ -267,21 +269,22 @@ void setstatus(char *what)
     tty_foreground(StatusForeground);
     tty_background(StatusBackground);
 
-    window_cursormove(status_win, 0, 0);
-    window_write(temp, SCREEN_X);	
+    window_cursormove_notify(status_win, 0, 0);
+    window_write(temp, SCREEN_X - (sizeof(signals[0].signame) - 1) - 1);
 }
 
 
 void setsignal(void)
 {
-    int len = strlen(signals[signal_type].signame);
+    int len = sizeof(signals[0].signame) - 1;
 
     tty_bright(StatusBrightness);
     tty_foreground(StatusForeground);
     tty_background(StatusBackground);
 
-    window_cursormove(status_win, 0, SCREEN_X - len - 1);
+    window_cursormove_notify(status_win, 0, SCREEN_X - len - 1);
     window_write(signals[signal_type].signame, len);
+    window_putch(' ');
 }
 
 
@@ -388,7 +391,7 @@ void update_process(int process, int update_color)
 	tty_bright(ScreenBrightness);
     }
 
-    window_cursormove(screen_win, process - first_on_screen, 0);
+    window_cursormove_notify(screen_win, process - first_on_screen, 0);
     window_write(temp, SCREEN_X);
 }
 
@@ -396,20 +399,33 @@ void update_process(int process, int update_color)
 void update_all(void)
 {
     int i;
+    char fill[256];
 
     tty_foreground(ScreenForeground);
     tty_background(ScreenBackground);
     tty_bright(ScreenBrightness);
     window_cursormove(screen_win, 0, 0);
+    memset(fill, ' ', SCREEN_X);
 
     for (i = first_on_screen;
          i < processes && (i - first_on_screen < SCREEN_Y - 3); i++)
 	    if (i != current_process)
 		update_process(i, OFF);
 	    else
-		window_cursormove(screen_win, i - first_on_screen + 1, 0);
+		window_cursormove_notify(screen_win, i - first_on_screen, 0);
 
     update_process(current_process, ON);
+
+    tty_background(ScreenBackground);
+    tty_bright(ScreenBrightness);
+
+    for (; i - first_on_screen < SCREEN_Y - 3; i++)
+    {
+	window_cursormove_notify(screen_win, i - first_on_screen, 0);
+	window_write(fill, SCREEN_X);
+    }
+
+    window_cursormove(screen_win, current_process-first_on_screen, SCREEN_X-1);
 }
 
 
@@ -425,6 +441,7 @@ void clean_up(void)
 
     tty_end();
     tty_clrscr();
+    tty_exit();
     removelog();
 }
 
@@ -433,7 +450,6 @@ void fatal(char *postmsg)
 {
     clean_up();
     fprintf(stderr, "Fatal error : %s !\n", postmsg);
-    removelog();
     exit(1);    
 }
 
@@ -451,9 +467,9 @@ int ps(char *args)
     stderr_log = fopen(fname, "w");
 
     if (args)
-	sprintf(ps_cmd, "ps %s", args);
+	sprintf(ps_cmd, "exec ps %s", args);
     else
-	sprintf(ps_cmd, "ps");
+	sprintf(ps_cmd, "exec ps");
 
     if (system(ps_cmd))
     {
@@ -503,12 +519,13 @@ void panic(int signum)
 
 	    tty_end();
 	    tty_clrscr();
+	    tty_exit();
+	    removelog();
 	    fprintf(stderr, "Got SIGSEGV !\n");
 	    fprintf(stderr, "Please report to tudor@ulise.cs.pub.ro\n");
 	    break;
     }
 
-    removelog();
     exit(1);
 }
 
@@ -519,7 +536,7 @@ int main(int argc, char *argv[])
     struct key_struct *ks;
     int key, repeat_count;
     char tmp[256], buf[256];
-    int sectionptr = 0, i, j, index;
+    int sectionptr = 0, i, j, index, first_time = 1;
     char *home, *data = NULL, fname[PATH_MAX], *section;
     int need_update, need_update_all, old_current_process;
 
@@ -540,8 +557,6 @@ int main(int argc, char *argv[])
     signal(SIGUSR2, SIG_IGN);
     signal(SIGTSTP, SIG_IGN);
     signal(SIGCONT, SIG_IGN);
-    signal(SIGTTIN, SIG_IGN);
-    signal(SIGTTOU, SIG_IGN);
     signal(SIGALRM, SIG_IGN);
     signal(SIGPIPE, SIG_IGN);
     signal(SIGFPE,  SIG_IGN);
@@ -631,7 +646,7 @@ int main(int argc, char *argv[])
 		        tty_key_list_insert(key_seq, (void *)j);
 		}
 		else
-		    fprintf(stderr, "Invalid builtin operation: %s\n",
+		    fprintf(stderr, "Invalid built-in operation: %s\n",
 			    contents);
 	    }
 
@@ -653,12 +668,13 @@ int main(int argc, char *argv[])
     }
 
     tty_getsize(&SCREEN_X, &SCREEN_Y);
+    tty_startup();
 
     header_text = (char *)malloc(SCREEN_X + 1);
 
     title_win  = window_init(0, 0, 	      1,	    SCREEN_X);
     header_win = window_init(0, 1, 	      1,	    SCREEN_X);
-    screen_win = window_init(0, 2,            SCREEN_Y - 2, SCREEN_X);
+    screen_win = window_init(0, 2,            SCREEN_Y - 3, SCREEN_X);
     status_win = window_init(0, SCREEN_Y - 1, 1,	    SCREEN_X);
 
     tty_init();
@@ -687,10 +703,14 @@ restart:
     build_ps_list(stdout_log);
     fclose(stdout_log);
 
-    tty_foreground(ScreenForeground);
-    tty_background(ScreenBackground);
-    tty_bright(ScreenBrightness);
-    tty_clrscr();
+    if (first_time)
+    {
+	tty_foreground(ScreenForeground);
+	tty_background(ScreenBackground);
+	tty_bright(ScreenBrightness);
+	tty_clrscr();
+	first_time = 0;
+    }
 
     settitle();
     setstatus(NULL);
@@ -851,6 +871,10 @@ restart:
 	    case BUILTIN_SIGCHLD: signal_type = 12; setsignal(); break;
 	    case BUILTIN_SIGCONT: signal_type = 13; setsignal(); break;
 
+	    case BUILTIN_HardRefresh:
+
+	    	first_time = 1;
+
 	    case BUILTIN_Refresh:
 
 	    	goto restart;
