@@ -1,6 +1,6 @@
 /*
     Simple window management file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -44,7 +44,13 @@ int window_write(char *str, int length)
 
 int window_putch(char c)
 {
-    return tty_write(&c, 1);
+    return tty_putch(c);
+}
+
+
+void window_cursormove_notify(window *this, int y, int x)
+{
+    tty_cursormove_notify(y + this->starty, x + this->startx);
 }
 
 
