@@ -1,6 +1,6 @@
 /*
     Simple window management header file
-    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
+    Copyright (c) 1993, 1994 Tudor Hulubei & Andrei Pitis
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -18,8 +18,8 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 
-#ifndef __WINDOW_H
-#define __WINDOW_H
+#ifndef _WINDOW_H
+#define _WINDOW_H
 
 
 typedef struct
@@ -33,7 +33,8 @@ typedef struct
 window *window_init(int startx, int starty, int rows, int columns);
 int  window_write(char *str, int length);
 int  window_putch(char c);
+void window_cursormove_notify(window *this, int y, int x);
 void window_cursormove(window *this, int y, int x);
 
 
-#endif
+#endif	/* _WINDOW_H */
